#ifndef LIB_SPRINGSYSTEM_H
#define LIB_SPRINGSYSTEM_H

#include <vector>
#include "math/vec.h"
#include "lib/parsystem.h"

class Spring {
public:
	//system
	bool		active;
	Particle	*par[2];		//edge vertex indices within Particle[] structure
	//physics
	float		restlen;	//resting length
	float		ks;			//damping of position difference
	float		kd;			//damping of velocity difference
};
VECTOR_FIX_MSVC50(Spring);

class SpringSystem {
private:
	void reset();

public:
	SpringSystem() {reset();}
	~SpringSystem() {shutdown();}

	bool init() {return true;}
	void shutdown() {}

	std::vector<Spring> spring;

	// dp = change in position from body B to body A = positionA - positionB
	// dv = change in velocity from body B to body A = velocityA - velocityB
	static vec3f computeSpringForce(const vec3f &dp, const vec3f &dv, float ks, float kd, float restlen);

	void applyForces(ParSystem &parSys);

	int createSpring();
};

#endif
