#include "lib/emitparsystem.h"

#define EMIT_PARTICLE_MAX		3000

void EmitParSystem::reset() {
	live = NULL;
	dead = NULL;
	parCount = 0;
	liveCount = 0;
}

bool EmitParSystem::init(ParSystem *parSys) {
	shutdown();

	this->parSys = parSys;

	return true;
}

void EmitParSystem::shutdown() {

	//delete all particles

	EmitParticle *next,*p;

	for (p = live; p; p = next) {
		next = p->next;
		parSys->removeParticle(p->par);
		delete p;
	}

	for (p = dead; p; p = next) {
		next = p->next;
		parSys->removeParticle(p->par);
		delete p;
	}

	//and clear the list pointers

	reset();
}

EmitParticle *EmitParSystem::createEmitParticle() {

	//grab or alloc a new one
	EmitParticle *p = NULL;
	if (dead) {
		p = dead;
		dead = dead->next;
		if (dead) dead->prev = NULL;
	} else {

		parCount++;
		if (parCount > EMIT_PARTICLE_MAX) return NULL;

		p = new EmitParticle();
		if (!p) return NULL;
	}

	//link it up
	if (!live) {
		live = p;
		p->prev = NULL;
		p->next = NULL;
	} else {
		live->prev = p;
		p->next = live;
		live = p;
		p->prev = NULL;
	}

	liveCount++;
	return p;
}

void EmitParSystem::removeEmitParticle(EmitParticle *p) {

	//unalloc resources
	parSys->removeParticle(p->par);

	//unlink 'er first:
	if (p->prev) p->prev->next = p->next;
	else live = p->next;
	if (p->next) p->next->prev = p->prev;

	//drop the live count
	liveCount--;

	//add to the dead chain for later use
	//this prevents excessive alloc/unalloc's
	//but, unlike a vector, this doesnt leave dead particles in the same array, slowing down our search times
	//and, unlike vectors, this doesnt shift around our pointers upon resize
	p->next = dead;
	if (dead) dead->prev = p;
	dead = p;
	p->prev = NULL;
}
