#include <stdio.h>
#include <assert.h>
#include <math.h>
#include <time.h>

#include <vector>
#include <iostream>

#include "system/gl.h"
#include "defs/defs.h"
#include "lib/objfile.h"

using namespace std;

VECTOR_FIX_MSVC50(triIndex_t)

//dont use this flag unless its with homework #2 OBJ files
//#define OFFSET_TO_CENTER

void ObjFile::reset() {
	flags = 0;
	list = 0;
	indexArrayBuffer = NULL;
	bbox = box3f(vec3f(fInf,fInf,fInf),vec3f(-fInf,-fInf,-fInf));
	memset(title, 0, sizeof(title));
}

void ObjFile::unload() {
	uncompile();

	vertex.clear();
	normal.clear();
	texCoord.clear();
	tri.clear();
	edge.clear();
	
	if (indexArrayBuffer) delete indexArrayBuffer;

	//unload the materials too
	for (int i = 0; i < (int)materials.size(); i++) {
		if (materials[i]) delete materials[i];
	}
	materials.clear();

	//and reset
	reset();
}

static char *nextEndLine(char *p, char *buffer) {
	while(*p) {
		if (*p == '\n' || *p == '\r') break;
		p++;
	}
	return p;
}

static char *nextNewLine(char *p, char *buffer) {
	while (*p) {
		if (*p != '\n' && *p != '\r') break;
		p++;
	}
	return p;
}

static char *nextSpace(char *p, char *buffer) {
	while (*p) {
		if (*p == ' ' || *p == '\t') break;
		p++;
	}
	return p;
}

static char *nextNonSpace(char *p, char *buffer) {
	while (*p) {
		if (*p != ' ' && *p != '\t') break;
		p++;
	}
	return p;
}

static char *nextSpaceOrEndLine(char *p, char *buffer) {
	while (*p) {
		if (*p == ' ' || *p == '\t' || *p == '\r' || *p == '\n') break;
		p++;
	}
	return p;
}

static char *readFloats(char *p, char *buffer, float *dest, int dim) {
	p = nextSpace(p, buffer);
	for (int i = 0; i < dim; i++) {
		char *start = p = nextNonSpace(p, buffer);
		char *end = p = nextSpaceOrEndLine(p, buffer);

		char token[256];
		int size = 255;
		int strSize = (int)(end - start);
		if (strSize < size) size = strSize;
		if (size < 1) break;
		memcpy(token, start, size);
		token[size] = 0;
		dest[i] = (float)atof(token);
	}
	return nextNewLine(nextEndLine(p, buffer), buffer);
}

static char *readInts(char *p, char *buffer, int *dest, int dim) {
	p = nextSpace(p, buffer);
	for (int i = 0; i < dim; i++) {
		char *start = p = nextNonSpace(p, buffer);
		char *end = p = nextSpaceOrEndLine(p, buffer);

		char token[256];
		int size = 255;
		int strSize = (int)(end - start);
		if (strSize < size) size = strSize;
		if (size < 1) break;
		memcpy(token, start, size);
		token[size] = 0;
		dest[i] = atoi(token);
	}
	return nextNewLine(nextEndLine(p, buffer), buffer);
}

static char *readString(char *p, char *buffer, string &dest) {
	p = nextSpace(p, buffer);

	char *start = p = nextNonSpace(p, buffer);
	char *end = p = nextSpaceOrEndLine(p, buffer);

	char token[256];
	int size = 255;
	int strSize = (int)(end - start);
	if (strSize < size) size = strSize;
	if (size >= 1) {
		memcpy(token, start, size);
		token[size] = 0;
		dest = token;
	} else {
		dest = "";
	}

	return nextNewLine(nextEndLine(p, buffer), buffer);
}

bool ObjFile::loadMaterialFile(const char *filename) {

	FILE *file = fopen(filename, "r");
	if (!file) {
		printf("unable to open the file %s\n", filename);
		return false;
	}

	//todo - condition test this
	fseek(file, 0, SEEK_END);
	int size = ftell(file);
	fseek(file, 0, SEEK_SET);

	char *buffer = (char *)malloc(size+1);
	if (!buffer) { 
		fclose(file);
		return false;
	}

	size = (int)fread(buffer, 1, size, file);
	fclose(file);
	buffer[size] = 0;	//force null term our buffer

	char *p = buffer;
	while(*p){
		if (*p == '\n' || *p == '\r') {
			continue;
		} else if (*p == '#') {	//comment
			p = nextNewLine(nextEndLine(p, buffer), buffer);
		} else if (!memcmp(p, "newmtl", 6)) {
			ObjFileMaterial *m = new ObjFileMaterial();
			assert(m);
			p = readString(p, buffer, m->name);
			materials.push_back(m);
		} else if (!memcmp(p, "Ka", 2)) {
			if (materials.size()) p = readFloats(p,buffer, materials[materials.size()-1]->Ka,3);
		} else if (!memcmp(p, "Kd", 2)) {
			if (materials.size()) p = readFloats(p,buffer, materials[materials.size()-1]->Kd,3);
		} else if (!memcmp(p, "Ks", 2)) {
			if (materials.size()) p = readFloats(p,buffer, materials[materials.size()-1]->Ks,3);
		} else if (!memcmp(p, "illum", 5)) {
			if (materials.size()) p = readInts(p,buffer, &materials[materials.size()-1]->illum, 1);
		} else if (!memcmp(p, "Ns", 2)) {
			if (materials.size()) p = readFloats(p,buffer, &materials[materials.size()-1]->Ns, 1);
		} else if (!memcmp(p, "map_Kd", 6)) {
			if (materials.size()) {
				ObjFileMaterial *m = materials[materials.size()-1];
				assert(m);
				p = readString(p, buffer, m->map_Kd);
				if (!m->map_Kd_tex.load2D(m->map_Kd.c_str())) {
					printf("unable to load requested material %s\n", m->map_Kd.c_str());
				}
			} else {
				printf("found a map_Kd without any materials present\n");
			}
		} else {
			printf("mtl file unknown char: %c\n", *p);
			free(buffer);
			return false;
		}
	}

	free(buffer);
	return true;
}

bool ObjFile::load(const char *filename, int loadflags) {
	//in case there's already something in this object
	unload();

//	printf("loading model %s...\n", filename);

	FILE *file = fopen(filename, "r");
	if (!file) {
		printf("unable to open the file %s\n", filename);
		return false;
	}

	//TODO - extract the directory and change to that directory

	//todo - condition test this
	fseek(file, 0, SEEK_END);
	int size = ftell(file);
	fseek(file, 0, SEEK_SET);

	char *buffer = (char *)malloc(size+1);
	if (!buffer) { 
		fclose(file);
		return false;
	}

	size = (int)fread(buffer, 1, size, file);
	fclose(file);
	buffer[size] = 0;	//force null term our buffer

	//holds our current polygon indices
	//used for breaking polys down into triangles
	vector<triIndex_t> poly;

	//assume we have both until we find none
	flags = bitflag(USE_TEXCOORDS) | bitflag(USE_NORMALS) | loadflags;

	//our currently selected material (with the usemtl command)
	int curMtlIndex = -1;

	char *p = buffer;
	while(*p){
		if (*p == '\n' || *p == '\r') {
			p = nextNewLine(nextEndLine(p, buffer), buffer);
		} else if (*p == '#') {	//comment
			p = nextNewLine(nextEndLine(p, buffer), buffer);
		} else if (*p == 'v') {

			if (p[1] == 't') {
				vec2f v(0,0);
				p = readFloats(p,buffer,v,2);
				texCoord.push_back(v);
			} else if (p[1] == 'n') {
				vec3f v(0,0,0);
				p = readFloats(p,buffer,v,3);
				normal.push_back(v);
			} else {
				vec3f v(0,0,0);
				p = readFloats(p,buffer,v,3);
				bbox.stretch(v);
				vertex.push_back(v);
			}
		} else if (*p == 'f') {
			p = nextSpace(p, buffer);

			for(;;) {
				char *start = p = nextNonSpace(p, buffer);
				char *end = p = nextSpaceOrEndLine(p, buffer);

				char token[32];
				int size = 31;
				int strSize = (int)(end - start);
				if (strSize < size) size = strSize;
				if (size < 1) break;
				memcpy(token, start, size);
				token[size] = 0;

				triIndex_t vi = {-1,-1,-1};

				sscanf(token, "%d/%d/%d ", &vi.v, &vi.t, &vi.n);

				//if we couldnt even read the vertex index then skip this one
				//maybe even print out an error or something
				if (vi.v == -1) break;

				//should we keep a flag for every texcoord index that says
				//'the texcoord here exists' ?
				//the Java3D OBJ file loader disregards all texcoords in a group
				//(specified by the 'g' field)
				//if any face leaves out a texcoord
				//(i.e. per group, the texcoord count and vertex count must match)
				if (vi.t == -1) {
					vi.t = 1;
					flags &= ~bitflag(USE_TEXCOORDS);
				}
				if (vi.n == -1) {
					vi.n = 1;
					flags &= ~bitflag(USE_NORMALS);
				}

				//subtract one because the obj file starts indexing at 1
				vi.v--;
				vi.t--;
				vi.n--;

				poly.push_back(vi);
			}
			p = nextNewLine(nextEndLine(p, buffer), buffer);

			if ((int)poly.size() < 3) {
				printf("found a face with only %d vertices!\n", poly.size());
			} else {
				//triangulate the polygon
				//TODO - add edge flags (glEdgeFlag style)
				bool high = false;	//false implies left++, true implies right--
				int triIndex[3];
				for (int left = 1, right = (int)poly.size() - 1; left != right; high = !high) {
					if (!high) {
						triIndex[0] = left - 1;
						triIndex[1] = left;
						triIndex[2] = right;
						left++;
					} else {
						triIndex[0] = right - 1;
						triIndex[1] = right;
						triIndex[2] = left - 1;
						right--;
					}

					//our new tri:
					tri_t t;

					//connect the dots
					t.p[0] = poly[triIndex[0]];
					t.p[1] = poly[triIndex[1]];
					t.p[2] = poly[triIndex[2]];

					//tag on the material
					t.mtlGroupIndex = curMtlIndex;

					//and put it on the stack
					tri.push_back(t);
				}
			}

			poly.clear();
		} else if (!memcmp(p, "mtllib", 6)) {
			string mtlFileName;
			p = readString(p, buffer, mtlFileName);

			//load the material file:
			loadMaterialFile(mtlFileName.c_str());

		} else if (*p == 'g' || *p == 'o' || *p == 's') {	//not yet
			p = nextNewLine(nextEndLine(p, buffer), buffer);
		} else if (!memcmp(p, "usemtl", 6)) {
			string mtl;
			p = readString(p, buffer, mtl);
			curMtlIndex = -1;
			for (int i = 0; i < (int)materials.size(); i++) {
				ObjFileMaterial *m = materials[i];
				assert(m);
				if (!strcmp(m->name.c_str(), mtl.c_str())) {
					curMtlIndex = i;
					break;
				}
			}
		} else {
			printf("got unhandled char %c %d at index %d of size %d\n", *p, *p, (int)(p - buffer), size);
			break;
		}
	}

	free(buffer);

	if (!texCoord.size()) {
		flags &= ~bitflag(USE_TEXCOORDS);
	}

	if (!normal.size()) {
		flags &= ~bitflag(USE_NORMALS);
	}

	//verify triangle index validity
	int i;
	for (i = 0; i < (int)tri.size(); i++) {
		for (int j = 0; j < 3; j++) {
		
			triIndex_t *t = &tri[i].p[j];
			
			if (t->v < 0 || t->v >= (int)vertex.size()) {
				printf("found an out of bounds vertex at tri %d point %d with index %d, max %d\n", i,j,t->v, vertex.size());
				unload();
				return false;
			}

			if (flags & bitflag(USE_TEXCOORDS)) {
				if (t->t < 0 || t->t >= (int)texCoord.size()) {
					printf("found an out of bounds texcoord index %d, max %d\n", t->t, texCoord.size());
					flags &= ~bitflag(USE_TEXCOORDS);
				}
			}

			if (flags & bitflag(USE_NORMALS)) {
				if (t->n < 0 || t->n >= (int)normal.size()) {
					printf("found an out of bounds normal index %d, max %d\n", t->n, normal.size());
					flags &= ~bitflag(USE_NORMALS);
				}
			}
		}
	}

	//remove any flags if we didnt find any instances of the elements

	if (!(flags & bitflag(USE_TEXCOORDS))) {
		texCoord.clear();
	}

	if (!(flags & bitflag(USE_NORMALS)) || (flags & bitflag(REGEN_NORMALS))) {
		normal.clear();

		printf("found no/invaild normals - generating...\n");

		/* normal generation */

		//use this for vertices with no normals specified
		//such vertices are ones for which every face that references them never
		//provides an associated normal
		//I think.

		normal.resize(vertex.size());

		for (i = 0; i < (int)normal.size(); i++) {
			normal[i] = vec3f(0,0,0);
		}
		for (i = 0; i < (int)tri.size(); i++) {
			//copy over vertex references
			tri[i].p[0].n = tri[i].p[0].v;
			tri[i].p[1].n = tri[i].p[1].v;
			tri[i].p[2].n = tri[i].p[2].v;
			vec3f n = vec3unitNormal(
				vertex[tri[i].p[0].v],
				vertex[tri[i].p[1].v],
				vertex[tri[i].p[2].v]);
			normal[tri[i].p[0].n] += n;
			normal[tri[i].p[1].n] += n;
			normal[tri[i].p[2].n] += n;
		}
		for (i = 0; i < (int)normal.size(); i++) {
			normal[i] = vec3unit(normal[i]);
		}

		/* end normal generation */

		printf("done generating normals.\n");

		flags |= bitflag(USE_NORMALS);
	}

	// generate edges
	if (loadflags & bitflag(GEN_EDGES)) {

		printf("generating edges...\n");
		//for every pair of faces
		clock_t startTime = clock();
		int lastSecond = 0;

		for (i = 0; i < (int)tri.size() - 1; i++) {

			//for complex models...
			{
				clock_t thisTime = clock();
				int second = (thisTime - startTime) / (10 * CLOCKS_PER_SEC);
				if (second != lastSecond) {
					lastSecond = second;
					printf(" %d of %d (%d%%) done \n", i, tri.size(), (i*100/(int)tri.size()) );
				}
			}

			for (int j = i + 1; j < (int)tri.size(); j++) {

				edge_t e;
				e.t[0] = i;
				e.t[1] = j;

				bool found = false;

				for (int k = 0; k < 3 && !found; k++) {
					int kk = (k + 1) % 3;
					for (int l = 0; l < 3; l++) {
						int ll = (l + 1) % 3;
						
						if (tri[i].p[kk].v == tri[j].p[l].v &&
							tri[i].p[k].v == tri[j].p[ll].v)
						{

							e.v[0] = k;
							e.v[1] = l;
							edge.push_back(e);

							found = true;

							break;
						}
					}
				}
			}
		}
		printf("done generating edges.\n");
	}

#ifdef OFFSET_TO_CENTER
	{
		vec3f center = (bbox.max + bbox.min) * 0.5f;
		for (i = 0; i < (int)vertex.size(); i++) {
			vertex[i] -= center;
		}
		bbox.max -= center;
		bbox.min -= center;
	}
#endif

	// generate planes
	for (i = 0; i < (int)tri.size(); i++) {
		tri[i].plane = vec3unitNormal(
			vertex[tri[i].p[0].v],
			vertex[tri[i].p[1].v],
			vertex[tri[i].p[2].v]);
		tri[i].plane.w = -dot((vec3f&)tri[i].plane, vertex[tri[i].p[0].v]);
	}
	
	//somethigns wrong with this file ...
	if (vertex.size() <= 0) return false;

	//calc bounding radius
	brad = vertex[0].lenSq();
	for (i = 1; i < (int)vertex.size(); i++) {
		float r = vertex[i].lenSq();
		if (r > brad) brad = r;
	}
	brad = (float)sqrt(brad);


//	printf("found %d vertices, %d normals, %d texcoords, %d faces\n",
//		vertex.size(), normal.size(), texCoord.size(), tri.size());
//	printf("radius %f bbox %f,%f,%f to %f,%f,%f\n",
//		brad, bbox.min.x, bbox.min.y, bbox.min.z,
//		bbox.max.x, bbox.max.y,bbox.max.z);
//	printf("loaded model %s\n", filename);

//	printf("compiling...\n");
	compile();
//	printf("done compiling.\n");

	strncpy(title, filename, sizeof(title));

	return true;
}

void ObjFile::renderImmediate(const vec3f *vtxs, bool useMtl) const {

	int mat = -2;

	bool begun = false;
	bool didvtx = false;

	if (!useMtl) {
		glBegin(GL_TRIANGLES);
	}

	for (int i = 0; i < (int)tri.size(); i++) {
		if (tri[i].mtlGroupIndex != mat && useMtl) {
			if (begun) {
				glEnd();
				begun = true;
			}

			//take advantage of the fact that all tri's of matching textures are consecutive
			//for the most part, all the tri instances of a single tex will be in one contiguous area
			mat = tri[i].mtlGroupIndex;
			if (mat < 0 || mat >= (int)materials.size()) {
				//only disable the current texture if we will eventually be re-enabling it
				//otherwise let the host program make due of what to do with the texturing
				if (materials.size()) {
					glDisable(GL_TEXTURE_2D);
				}
			} else {
				assert(materials[mat]);
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, materials[mat]->map_Kd_tex.getTextureID());
			}

			//clear our flag as to whether or not we placed a vertex
			didvtx = false;
			glBegin(GL_TRIANGLES);
		}

		for (int j = 0; j < 3; j++) {
			if (flags & bitflag(USE_TEXCOORDS)) {
				glTexCoord2fv(texCoord[tri[i].p[j].t]);
			}

			if (flags & bitflag(USE_NORMALS)) {
				glNormal3fv(normal[tri[i].p[j].n]);
			}

			glVertex3fv(vtxs[tri[i].p[j].v]);
			didvtx = true;
		}
	}

	if (didvtx) glEnd();
}

void ObjFile::render() const {

	if (list) {
		glCallList(list);
	} else {
		renderImmediate();
	}
}

void ObjFile::compile() {
	//no can do with more than one texture ... unless we make a list per-texture ...
	//i'll have to work on that ...
	//until then ...
	if (!list) {
		list = glGenLists(1);
		glNewList(list, GL_COMPILE);
		renderImmediate();
		glEndList();
	}
}

void ObjFile::uncompile() {
	if (list) {
		//glDeleteLists(lists, materials.size()+1);
		glDeleteLists(list, 1);
		list = 0;
	}
}

void ObjFile::renderExtendNormals(float magnitude) {
	vec3f v;
	
	if (!(flags & bitflag(USE_NORMALS))) {
		printf("no normals found\n");
	}

	glBegin(GL_TRIANGLES);
	for (int i = 0; i < (int)tri.size(); i++) {
		for (int j = 0; j < 3; j++) {

			if (flags & bitflag(USE_TEXCOORDS)) {
				glTexCoord2fv(texCoord[tri[i].p[j].t]);
			}

			if (flags & bitflag(USE_NORMALS)) {
				glNormal3fv(normal[tri[i].p[j].n]);
				v = vertex[tri[i].p[j].v] + normal[tri[i].p[j].n] * magnitude;
				glVertex3fv(v);
			} else {
				glVertex3fv(vertex[tri[i].p[j].v]);
			}
		}
	}
	glEnd();
}

void ObjFile::renderWithArrays() {
	//come to think of it this only works when ... well ... 
	//it only works for special objects
	//like the 1-1 vertex to normal mapped objects i was working with ...
	//i should -- if i were p.c. -- i should keep a special flag to state if the obj was made with that constraint
	//but i'm gonna wing it instead.
	
	//are we all sure the compiler will 4-byte align these vectors?
	//cuz if its any more than that then the byte alignment should be calculated ...
	if (flags & bitflag(USE_TEXCOORDS)) {
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer(2, GL_FLOAT, 0, &(texCoord[0].x));
	}
	
	if (flags & bitflag(USE_NORMALS)) {
		glEnableClientState(GL_NORMAL_ARRAY);
		glNormalPointer(GL_FLOAT, 0, &(normal[0].x));
	}
	
	glEnableClientState(GL_VERTEX_ARRAY);
	glVertexPointer(3, GL_FLOAT, 0, &(vertex[0].x));
	
	if (!indexArrayBuffer) {
		//only fill it if we need it
		//gosh i could allocate this on the GPU if i was really tricky .. pixel buffer objects all the way!
		indexArrayBuffer = new GLuint[tri.size() * 3];
		assert(indexArrayBuffer);
		//the only reason i need a separate buffer is cuz i have all the tri stuff struct'd together ...
		//if i separated the vertices i'd be safe ...
		for (int i = 0; i < tri.size(); i++) {
			for (int j = 0; j < 3; j++) {
				int k = 3*i+j;
				indexArrayBuffer[k] = tri[i].p[j].v;
			}
		}
	}
	
	glDrawElements(GL_TRIANGLES, tri.size() * 3, GL_UNSIGNED_INT, indexArrayBuffer);
	
	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_NORMAL_ARRAY);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
}

