#include <assert.h>
#include "lib/springobj.h"

void SpringObj::clear() {
	vertex.clear();
	spring.clear();
	parSys = NULL;
	springSys = NULL;
}

bool SpringObj::init(ParSystem *parSys, SpringSystem *springSys) {
	//shutdown any existing data:
	shutdown();

	this->parSys = parSys;
	this->springSys = springSys;

	return true;
}

void SpringObj::shutdown() {

	if (parSys) {
		for (int i = 0; i < (int)vertex.size(); i++) {
			if (vertex[i]) {
				parSys->removeParticle(vertex[i]);
			}
		}
	}

	if (springSys) {
		for (int i = 0; i < (int)spring.size(); i++) {
			if (spring[i] >= 0 && spring[i] < (int)springSys->spring.size()) {
				springSys->spring[spring[i]].active = false;
			}
		}
	}

	//return all variables to their defaults:
	clear();
}

void SpringObj::rotate(const quatf &q) {
	for (int i = 0; i < (int)vertex.size(); i++) {
		assert(vertex[i]);
		Particle *p = vertex[i];
		p->x = quatRotate(q, p->x);
	}
}

void SpringObj::translate(const vec3f &t) {
	for (int i = 0; i < (int)vertex.size(); i++) {
		assert(vertex[i]);
		Particle *p = vertex[i];
		p->x += t;
	}
}

void SpringObj::scale(const vec3f &s) {
	for (int i = 0; i < (int)vertex.size(); i++) {
		assert(vertex[i]);
		Particle *p = vertex[i];
		p->x.x *= s.x;
		p->x.y *= s.y;
		p->x.z *= s.z;
	}
}

void SpringObj::resetSpringLengths() {
	for (int i = 0; i < (int)spring.size(); i++) {
		assert(spring[i] >= 0 && spring[i] < (int)springSys->spring.size());
		Spring &s = springSys->spring[spring[i]];
		assert(s.active);
		if (!s.par[0] || !s.par[1]) {
			printf("spring %d has bad particle poitners\n", spring[i]);
			s.restlen = 0;
		} else {
			s.restlen = (float)(s.par[0]->x - s.par[1]->x).len();
		}
	}
}

void SpringObj::setSpringKs(float ks) {
	for (int i = 0; i < (int)spring.size(); i++) {
		assert(spring[i] >= 0 && spring[i] < (int)springSys->spring.size());
		Spring &s = springSys->spring[spring[i]];
		assert(s.active);
		s.ks = ks;
	}
}

void SpringObj::setSpringKd(float kd) {
	for (int i = 0; i < (int)spring.size(); i++) {
		assert(spring[i] >= 0 && spring[i] < (int)springSys->spring.size());
		Spring &s = springSys->spring[spring[i]];
		assert(s.active);
		s.kd = kd;
	}
}
